"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const lsp_editor_adapter_1 = require("lsp-editor-adapter");
const coreutils_1 = require("@jupyterlab/coreutils");
const codemirror_1 = require("./codemirror");
const converter_1 = require("../converter");
const free_tooltip_1 = require("../free_tooltip");
const widgets_1 = require("@phosphor/widgets");
const utils_1 = require("../utils");
const signaling_1 = require("@phosphor/signaling");
/**
 * Foreign code: low level adapter is not aware of the presence of foreign languages;
 * it operates on the virtual document and must not attempt to infer the language dependencies
 * as this would make the logic of inspections caching impossible to maintain, thus the WidgetAdapter
 * has to handle that, keeping multiple connections and multiple virtual documents.
 */
class JupyterLabWidgetAdapter {
    constructor(app, rendermime_registry, invoke) {
        this.app = app;
        this.rendermime_registry = rendermime_registry;
        this.invoke_command = invoke;
        this.connections = new Map();
        this.document_connected = new signaling_1.Signal(this);
    }
    get root_path() {
        // TODO: serverRoot may need to be included for Hub or Windows, requires testing.
        // let root = PageConfig.getOption('serverRoot');
        return coreutils_1.PathExt.dirname(this.document_path);
    }
    invoke_completer() {
        return this.app.commands.execute(this.invoke_command);
    }
    get main_connection() {
        return this.connections.get(this.virtual_editor.virtual_document.id_path);
    }
    connect(virtual_document) {
        return __awaiter(this, void 0, void 0, function* () {
            virtual_document.foreign_document_opened.connect((host, context) => {
                this.connect(context.foreign_document);
            });
            virtual_document.foreign_document_closed.connect((host, context) => {
                let connection = this.connections.get(context.foreign_document.id_path);
                connection.close();
            });
            let language = virtual_document.language;
            console.log(`LSP: will connect using root path: ${this.root_path} and language: ${language}`);
            let connection = new lsp_editor_adapter_1.LspWsConnection({
                serverUri: 'ws://localhost/' + language,
                languageId: language,
                // paths handling needs testing on Windows and with other language servers
                rootUri: 'file:///' + this.root_path,
                documentUri: 'file:///' + virtual_document.uri,
                documentText: () => {
                    // TODO VirtualEditor should have "update(force=false)" method with debouncer;
                    //  current getValue() should be moved to update(), while getValue should be
                    //  made private in the virtual editor; possibly the virtual editor should no
                    //  longer expose the full implementation of CodeMirror but rather hide it inside.
                    this.virtual_editor.getValue();
                    try {
                        return virtual_document.value;
                    }
                    catch (e) {
                        console.log(e);
                    }
                }
            }).connect(new WebSocket('ws://localhost:3000/' + language));
            // @ts-ignore
            connection.on('goTo', locations => this.handle_jump(locations, language));
            this.connections.set(virtual_document.id_path, connection);
            // @ts-ignore
            yield utils_1.until_ready(() => connection.isConnected, -1, 150);
            console.log('LSP:', this.document_path, 'connected.');
            this.create_adapter(virtual_document, connection);
            this.document_connected.emit({ document: virtual_document, connection: connection });
        });
    }
    handle_jump(locations, language) {
        let connection = this.connections.get(language);
        // TODO: implement selector for multiple locations
        //  (like when there are multiple definitions or usages)
        if (locations.length === 0) {
            console.log('No jump targets found');
            return;
        }
        console.log('Will jump to the first of suggested locations:', locations);
        let location = locations[0];
        let uri = decodeURI(location.uri);
        let current_uri = connection.getDocumentUri();
        let cm_position = converter_1.PositionConverter.lsp_to_cm(location.range.start);
        let editor_index = this.virtual_editor.get_editor_index(cm_position);
        let transformed_position = this.virtual_editor.transform_virtual_to_source(cm_position);
        let transformed_ce_position = converter_1.PositionConverter.cm_to_ce(transformed_position);
        console.log(`Jumping to ${transformed_position} in ${editor_index} editor of ${uri}`);
        if (uri === current_uri) {
            this.jumper.jump({
                token: {
                    offset: this.jumper.getOffset(transformed_ce_position, editor_index),
                    value: ''
                },
                index: editor_index
            });
            return;
        }
        if (uri.startsWith('file://')) {
            uri = uri.slice(7);
        }
        this.jumper.global_jump({
            // TODO: there are many files which are not symlinks
            uri: '.lsp_symlink/' + uri,
            editor_index: editor_index,
            line: transformed_ce_position.line,
            column: transformed_ce_position.column
        }, true);
    }
    create_adapter(virtual_document, connection) {
        this.adapter = new codemirror_1.CodeMirrorAdapterExtension(connection, { quickSuggestionsDelay: 50 }, this.virtual_editor, this.create_tooltip.bind(this), this.invoke_completer.bind(this), virtual_document);
        console.log('LSP: Adapter for', this.document_path, 'is ready.');
    }
    get_doc_position_from_context_menu() {
        // get the first node as it gives the most accurate approximation
        let leaf_node = this.app.contextMenuHitTest(() => true);
        let { left, top } = leaf_node.getBoundingClientRect();
        // @ts-ignore
        let event = this.app._contextMenuEvent;
        // if possible, use more accurate position from the actual event
        // (but this relies on an undocumented and unstable feature)
        if (event !== undefined) {
            left = event.clientX;
            top = event.clientY;
            event.stopPropagation();
        }
        return this.virtual_editor.coordsChar({
            left: left,
            top: top
        }, 'window');
    }
    create_tooltip(markup, cm_editor, position) {
        const bundle = markup.kind === 'plaintext'
            ? { 'text/plain': markup.value }
            : { 'text/markdown': markup.value };
        const tooltip = new free_tooltip_1.FreeTooltip({
            anchor: this.widget.content,
            bundle: bundle,
            editor: this.find_ce_editor(cm_editor),
            rendermime: this.rendermime_registry,
            position: converter_1.PositionConverter.cm_to_ce(position),
            moveToLineEnd: false
        });
        widgets_1.Widget.attach(tooltip, document.body);
        return tooltip;
    }
}
exports.JupyterLabWidgetAdapter = JupyterLabWidgetAdapter;
