"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const jupyterlab_1 = require("./jupyterlab");
const notebook_1 = require("../virtual/editors/notebook");
const utils_1 = require("../utils");
const completion_1 = require("../completion");
const regexp_1 = require("../extractors/regexp");
const defaults_1 = require("../magics/defaults");
// TODO: make the regex code extractors configurable
let foreign_code_extractors = {
    // general note: to match new lines use [^] instead of dot, unless the target is ES2018, then use /s
    python: [
        // R magic will always be in the same, single R-namespace
        new regexp_1.RegExpForeignCodeExtractor({
            language: 'R',
            pattern: '^%%R( .*?)?\n([^]*)',
            extract_to_foreign: '$2',
            keep_in_host: true,
            is_standalone: false
        }),
        // R line magic
        new regexp_1.RegExpForeignCodeExtractor({
            language: 'R',
            pattern: '(^|\n)%R (.*)\n?',
            extract_to_foreign: '$2',
            keep_in_host: true,
            is_standalone: false
        }),
        // script magics are standalone, i.e. consecutive code cells with the same magic create two different namespaces
        new regexp_1.RegExpForeignCodeExtractor({
            language: 'python',
            pattern: '^%%(python|python2|python3)( .*?)?\n([^]*)',
            extract_to_foreign: '$3',
            keep_in_host: false,
            is_standalone: true
        }),
        new regexp_1.RegExpForeignCodeExtractor({
            language: 'perl',
            pattern: '^%%(perl)( .*?)?\n([^]*)',
            extract_to_foreign: '$3',
            keep_in_host: false,
            is_standalone: true
        }),
        // js is not a script magic
        new regexp_1.RegExpForeignCodeExtractor({
            language: 'js',
            pattern: '^%%(js|javascript)( .*?)?\n([^]*)',
            extract_to_foreign: '$3',
            keep_in_host: false,
            is_standalone: false
        })
    ]
};
class NotebookAdapter extends jupyterlab_1.JupyterLabWidgetAdapter {
    constructor(editor_widget, jumper, app, completion_manager, rendermime_registry) {
        super(app, rendermime_registry, 'completer:invoke-notebook');
        this.widget = editor_widget;
        this.editor = editor_widget.content;
        this.completion_manager = completion_manager;
        this.jumper = jumper;
        this.init_once_ready().then();
    }
    is_ready() {
        return (this.widget.context.isReady &&
            this.widget.content.isVisible &&
            this.widget.content.widgets.length > 0 &&
            this.language !== '');
    }
    get document_path() {
        return this.widget.context.path;
    }
    get language() {
        let language_metadata = this.widget.model.metadata.get('language_info');
        // @ts-ignore
        return language_metadata.name;
    }
    find_ce_editor(cm_editor) {
        return this.virtual_editor.cm_editor_to_cell.get(cm_editor).editor;
    }
    init_once_ready() {
        return __awaiter(this, void 0, void 0, function* () {
            console.log('LSP: waiting for', this.document_path, 'to fully load');
            yield utils_1.until_ready(this.is_ready.bind(this), -1);
            console.log('LSP:', this.document_path, 'ready for connection');
            // TODO
            // this.widget.context.pathChanged
            this.virtual_editor = new notebook_1.VirtualEditorForNotebook(this.widget, this.language, defaults_1.language_specific_overrides, foreign_code_extractors, this.document_path);
            this.connect(this.virtual_editor.virtual_document);
            // refresh server held state after every change
            // note this may be changed soon: https://github.com/jupyterlab/jupyterlab/issues/5382#issuecomment-515643504
            this.widget.model.contentChanged.connect(this.refresh_lsp_notebook_image.bind(this));
            // and refresh it after the cell was activated, just to make sure that the first experience is ok
            this.widget.content.activeCellChanged.connect(this.refresh_lsp_notebook_image.bind(this));
            // register completion connectors on cells
            this.document_connected.connect(() => this.connect_completion());
        });
    }
    connect(virtual_document) {
        const _super = Object.create(null, {
            connect: { get: () => super.connect }
        });
        return __awaiter(this, void 0, void 0, function* () {
            return _super.connect.call(this, virtual_document);
        });
    }
    connect_completion() {
        // see https://github.com/jupyterlab/jupyterlab/blob/c0e9eb94668832d1208ad3b00a9791ef181eca4c/packages/completer-extension/src/index.ts#L198-L213
        const cell = this.widget.content.activeCell;
        const connector = new completion_1.LSPConnector({
            editor: cell.editor,
            connections: this.connections,
            virtual_editor: this.virtual_editor,
            session: this.widget.session
        });
        const handler = this.completion_manager.register({
            connector,
            editor: cell.editor,
            parent: this.widget
        });
        this.widget.content.activeCellChanged.connect((notebook, cell) => {
            const connector = new completion_1.LSPConnector({
                editor: cell.editor,
                connections: this.connections,
                virtual_editor: this.virtual_editor,
                session: this.widget.session
            });
            handler.editor = cell.editor;
            handler.connector = connector;
        });
    }
    get_document_content() {
        return this.virtual_editor.getValue();
    }
    refresh_lsp_notebook_image(slot) {
        // TODO this is fired too often currently, debounce!
        this.main_connection.sendChange();
    }
}
exports.NotebookAdapter = NotebookAdapter;
